/* $Id: tenm_graphic.c,v 1.84 2002/07/10 06:35:17 oohara Exp $ */
#include <stdio.h>
/* atexit, malloc  */
#include <stdlib.h>
#include <math.h>
/* for tenm_load_image */
#include <stdarg.h>
/* isprint */
#include <ctype.h>

#include <SDL.h>
#include <SDL_image.h>

#include "tenm_math.h"
#include "tenm_primitive.h"
#include "tenm_object.h"
#include "tenm_sdl_init.h"

#include "tenm_graphic.h"

/* number of printable characters */
#define TENM_FONT_NUMBER_CHAR 95

static SDL_Surface *dpy = NULL;

static int tenm_font_index(int c);

/* return 0 on success, 1 on error */
int
tenm_graphic_init(int width, int height)
{
  if (SDL_WasInit(SDL_INIT_VIDEO) != 0)
  {
    fprintf(stderr, "tenm_graphic_init: warning: video is already "
            "initialized\n");
    return 0;
  }

  if (tenm_sdl_init(SDL_INIT_VIDEO) != 0)
  {
    fprintf(stderr, "tenm_graphic_init: cannot initialize SDL video \n");
    return 1;
  }

  if (dpy != NULL)
  {
    fprintf(stderr, "tenm_graphic_init: "
            "dpy is already set\n");
    /* this is a fatal error --- if we are here, video is not
     * initialized before tenm_graphic_init() is called, so
     * the situation is weird enough
     */
    /* dpy should not be freed here --- it is the job of SDL_Quit() */
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
    return 1;
  }
  dpy = SDL_SetVideoMode(width, height, 0, 0);
  if (dpy == NULL)
  {
    fprintf(stderr, "tenm_graphic_init: "
            "SDL_SetVideoMode failed (%s)\n", SDL_GetError());
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
    return 1;
  }

  SDL_WM_SetCaption("tenmado", "tenmado");
  SDL_ShowCursor(SDL_DISABLE);

  return 0;
}

/* draw a line between A(a_x, a_y) and B(b_x, b_y)
 * return 0 on success, 1 on error
 */
int
tenm_draw_line(int a_x, int a_y, int b_x, int b_y, int width,
               tenm_color color)
{
  int i;
  int j;
  int a_x_temp;
  int a_y_temp;
  int b_x_temp;
  int b_y_temp;
  int status = 0;
  double lean;
  double vertical_x;
  double vertical_y;
  double vertical_length;
  int src_x;
  int src_y;

  /* sanity check */
  if (width <= 0)
    return 0;

  if (a_x <= b_x)
  {
    a_x_temp = a_x;
    b_x_temp = b_x;
  }
  else
  {
    a_x_temp = b_x;
    b_x_temp = a_x;
  }
  if (a_y <= b_y)
  {
    a_y_temp = a_y;
    b_y_temp = b_y;
  }
  else
  {
    a_y_temp = b_y;
    b_y_temp = a_y;
  }

  if (a_x == b_x)
  {
    for (i = a_y_temp; i <= b_y_temp; i++)
      for (j = a_x_temp - (width - 1); j <= b_x_temp + (width - 1); j++)
        if (tenm_draw_point(j, i, color) != 0)
          status = 1;
    return status;
  }
  
  if (a_y == b_y)
  {
    for (i = a_x_temp; i <= b_x_temp; i++)
      for (j = a_y_temp - (width - 1); j <= b_y_temp + (width - 1); j++)
        if (tenm_draw_point(i, j, color) != 0)
          status = 1;
    return status;
  }

  /* if we reached here, a_x != b_x and a_y != b_y */
  vertical_length = tenm_sqrt ((b_y - a_y) * (b_y - a_y)
                               + (a_x - b_x) * (a_x - b_x));
  vertical_x = ((double) (b_y - a_y)) / vertical_length;
  vertical_y = ((double) (a_x - b_x)) / vertical_length;

  if (b_x_temp - a_x_temp >=  b_y_temp - a_y_temp)
  {
    lean = ((double) (b_y - a_y)) / ((double) (b_x - a_x));
    for (i = a_x_temp; i <= b_x_temp; i++)
    {
      src_y = a_y + (int) (lean * (double) (i - a_x));
      for (j = - (width - 1); j <= width -1; j++)
        if (tenm_draw_point(i + (int) (vertical_x * (double) j),
                            src_y + (int) (vertical_y * (double) j),
                            color) != 0)
          status = 1;
    }
  }
  else
  {
    lean = ((double) (b_x - a_x)) / ((double) (b_y - a_y));
    for (i = a_y_temp; i <= b_y_temp; i++)
    {
      src_x = a_x + (int) (lean * (double) (i - a_y));
      for (j = - (width - 1); j <= width -1; j++)
        if (tenm_draw_point(src_x + (int) (vertical_x * (double) j),
                            i + (int) (vertical_y * (double) j),
                            color) != 0)
          status = 1;
    }
  }

  return status;
}

/* draw a circle whose center is (center_x, center_y)
 * return 0 on success, 1 on error
 */
int
tenm_draw_circle(int center_x, int center_y, int radius, tenm_color color)
{
  int i;
  int j;
  int status = 0;
  int temp;
  int temp2 = 0;
  int temp3;
  int temp4;

  /* sanity check */
  if (radius <= 0)
    return 0;

  temp = radius * radius;
  temp2 = (int) tenm_sqrt(2 * temp2);
  /*
  for (i = 0; i < 360; i++)
  {
    if (tenm_draw_point(center_x + (int) (tenm_cos(i) * (double) radius),
                        center_y + (int) (tenm_sin(i) * (double) radius),
                        color) != 0)
      status = 1;
  }
  */
  for (i = radius; i >= temp2; i--)
  {
    temp3 = (int) tenm_sqrt(temp - i * i);
    temp4 = (int) tenm_sqrt(temp - (i - 1) * (i - 1));
    for (j = temp3; j < temp4; j++)
    {
      if (tenm_draw_point(center_x + i,
                          center_y + j,
                          color) != 0)
        status = 1;
      if (tenm_draw_point(center_x + i,
                          center_y - j,
                          color) != 0)
        status = 1;
      if (tenm_draw_point(center_x - i,
                          center_y + j,
                          color) != 0)
        status = 1;
      if (tenm_draw_point(center_x - i,
                          center_y - j,
                          color) != 0)
        status = 1;

      if (tenm_draw_point(center_x + j,
                          center_y + i,
                          color) != 0)
        status = 1;
      if (tenm_draw_point(center_x + j,
                          center_y - i,
                          color) != 0)
        status = 1;
      if (tenm_draw_point(center_x - j,
                          center_y + i,
                          color) != 0)
        status = 1;
      if (tenm_draw_point(center_x - j,
                          center_y - i,
                          color) != 0)
        status = 1;
    }
  }

  return status;
}

/* you must _NOT_ lock dpy before calling this function
 * return 0 on success, 1 on error
 */
int
tenm_draw_point(int x, int y, tenm_color color)
{
  return tenm_draw_rectangle(x, y, 1, 1, color);
}

/* you must _NOT_ lock dpy before calling this function
 * return 0 on success, 1 on error
 */
int
tenm_draw_rectangle(int x, int y, int w, int h, tenm_color color)
{
  SDL_Rect rec;

  /* sanity check */
  if ((x + (w - 1) < 0) || (x - (w - 1) >= dpy->w))
    return 0;
  if ((y + (h - 1) < 0) || (y - (h - 1) >= dpy->h))
    return 0;

  rec.x = x;
  rec.y = y;
  rec.w = w;
  rec.h = h;
  if (SDL_FillRect(dpy, &rec, color) != 0)
    return 1;
  return 0;
}

tenm_color
tenm_map_color(int red, int green, int blue)
{
  return SDL_MapRGB(dpy->format, (Uint8) red, (Uint8) green, (Uint8) blue);
}

/* return 0 on success, 1 on error */
int
tenm_clear_window(tenm_color color)
{
  if (SDL_FillRect(dpy, NULL, color) != 0)
  {
    fprintf(stderr, "tenm_clear_window: SDL_FillRect failed (%s)\n",
            SDL_GetError());
    return 1;
  }
  return 0;
}

/* return 0 on success, 1 on error */
int
tenm_redraw_window(void)
{
  if (SDL_Flip(dpy) != 0)
  {
    fprintf(stderr, "tenm_redraw_window: SDL_Flip failed (%s)\n",
            SDL_GetError());
    return 1;
  }
  return 0;
}

/* return 0 on success, 1 on error */
int
tenm_draw_primitive(tenm_primitive *p, tenm_color color)
{
  int i;
  int j;
  int status = 0;

  if (p == NULL)
  {
    fprintf(stderr, "tenm_draw_primitive: p is NULL\n");
    return 0;
  }

  switch (p->class)
  {
  case TENM_POINT:
    if (tenm_draw_point((int) ((tenm_point *) p)->x,
                        (int) ((tenm_point *) p)->y,
                        color) != 0)
      return 1;
    else
      return 0;
    break;
  case TENM_CIRCLE:
    if (tenm_draw_circle((int) ((tenm_circle *) p)->center->x,
                         (int) ((tenm_circle *) p)->center->y,
                         (int) ((tenm_circle *) p)->r,
                         color) != 0)
      return 1;
    else
      return 0;
    break;
  case TENM_SEGMENT:
    if (tenm_draw_line((int) ((tenm_segment *) p)->a->x,
                       (int) ((tenm_segment *) p)->a->y,
                       (int) ((tenm_segment *) p)->b->x,
                       (int) ((tenm_segment *) p)->b->y,
                       1, color) != 0)
      return 1;
    else
      return 0;
    break;
  case TENM_POLYGON:
    for (i = 0; i < ((tenm_polygon *) p)->n; i++)
      for (j = i + 1; j < ((tenm_polygon *) p)->n; j++)
        if (tenm_draw_line((int) ((tenm_polygon *) p)->v[i]->x,
                           (int) ((tenm_polygon *) p)->v[i]->y,
                           (int) ((tenm_polygon *) p)->v[j]->x,
                           (int) ((tenm_polygon *) p)->v[j]->y,
                           1, color) != 0)
          status = 1;
    return status;
    break;
  default:
    fprintf(stderr, "tenm_draw_primitive: strange primitive (%d)\n", p->class);
    return 1;
    break;
  }
  /* should not reach here */
  return 1;
}

/* return 0 on success, 1 on error */
int
tenm_draw_mass(tenm_mass *p, tenm_color color)
{
  int i;
  int status = 0;

  /* sanity check */
  if (p == NULL)
    return 0;

  for (i = 0; i < p->n; i++)
    if (tenm_draw_primitive(p->p[i], color))
      status = 1;
  return status;
}

/* load image files and split them
 * ... is one or more of
 * (const char *file_name, int row, int column, tenm_color transparent_color)
 * return NULL on error
 */
tenm_image *
tenm_load_image(int number_file, ...)
{
  int i;
  int j;
  int k;
  int l;
  const char *file_name;
  int row;
  int column;
  tenm_color transparent_color;
  int width;
  int height;
  int suffix;
  va_list ap;
  tenm_image *temp;
  SDL_Surface **temp_surface;
  SDL_Surface *master;
  SDL_Rect temp_rect_src;
  SDL_Rect temp_rect_dest;

  /* sanity check */
  if (number_file <= 0)
  {
    fprintf(stderr, "tenm_load_image: strange number_file (%d)\n",
            number_file);
    return NULL;
  }

  temp = (tenm_image *) malloc(sizeof(tenm_image));
  if (temp == NULL)
  {
    fprintf(stderr, "tenm_load_image: malloc failed\n");
    return NULL;
  }
  temp->n = 0;
  temp->image = NULL;

  va_start(ap, number_file);
  for (i = 0; i < number_file; i++)
  {
    /* you can't use char in va_arg */
    file_name = (const char *) va_arg(ap, const int *);
    row = va_arg(ap, int);
    column = va_arg(ap, int);
    transparent_color = va_arg(ap, tenm_color);

    /* sanity check */
    if (row <= 0)
      continue;
    if (column <= 0)
      continue;

    master = IMG_Load(file_name);
    if (master == NULL)
    {
      fprintf(stderr, "tenm_load_image: IMG_Load(%s) failed\n", file_name);
      if (temp->image != NULL)
      {
        for (l = 0; l < temp->n; l++)
          SDL_FreeSurface(temp->image[l]);
        free(temp->image);
      }
      free(temp);
      va_end(ap);
      return NULL;
    }
 
    width = master->w / column;
    if (master->w % column != 0)
      fprintf(stderr, "tenm_load_image: warning: width (%d) of image (%s) "
              "cannot be divided by %d\n", master->w, file_name, column);
    height = master->h / row;
    if (master->h % row != 0)
      fprintf(stderr, "tenm_load_image: warning: height (%d) of image (%s) "
              "cannot be divided by %d\n", master->h, file_name, height);

    /* don't substitute tenm->image directly, or you will be in a trouble
     * if realloc fails  */
    if (temp->image == NULL)
      temp_surface = (SDL_Surface **) malloc(sizeof(SDL_Surface *)
                                             * row * column);
    else
      temp_surface = (SDL_Surface **) realloc(temp->image,
                                              sizeof(SDL_Surface *)
                                              * (row * column + temp->n));
    if (temp_surface == NULL)
    {
      fprintf(stderr, "tenm_load_image: memory allocation to image "
              "(%d) failed\n", i);
      SDL_FreeSurface(master);
      if (temp->image != NULL)
      {
        for (l = 0; l < temp->n; l++)
          SDL_FreeSurface(temp->image[l]);
        free(temp->image);
      }
      free(temp);
      return NULL;
    }
    temp->image = temp_surface;

    for (j = 0; j < row; j++)
      for (k = 0; k < column; k++)
      {
        suffix = temp->n + j * column + k;
        /* plain malloc is not enough */
        temp->image[suffix] = SDL_CreateRGBSurface(0, width, height,
                                                   dpy->format->BitsPerPixel,
                                                   dpy->format->Rmask,
                                                   dpy->format->Gmask,
                                                   dpy->format->Bmask,
                                                   dpy->format->Amask);
        if (temp->image[suffix] == NULL)
        {
          fprintf(stderr, "tenm_load_image: SDL_CreateRGBSurface(%d) "
                  "failed (%s)\n", suffix, SDL_GetError());
          SDL_FreeSurface(master);
          for (l = 0; l < suffix; l++)
            SDL_FreeSurface(temp->image[l]);
          free(temp->image);
          free(temp);
          return NULL;
        }
        if (dpy->format->BitsPerPixel == 8)
          SDL_SetPalette(temp->image[suffix], SDL_LOGPAL|SDL_PHYSPAL,
                         dpy->format->palette->colors,
                         0, dpy->format->palette->ncolors);
        temp_rect_src.x = width * k;
        temp_rect_src.y = row * j;
        temp_rect_src.w = width;
        temp_rect_src.h = height;
        temp_rect_dest.x = 0;
        temp_rect_dest.y = 0;
        temp_rect_dest.w = width;
        temp_rect_dest.h = height;
        if (SDL_BlitSurface(master, &temp_rect_src,
                            temp->image[suffix], &temp_rect_dest) != 0)
        {
          fprintf(stderr, "tenm_load_image: SDL_BlitSurface failed (%s)\n",
                  SDL_GetError());
          SDL_FreeSurface(master);
          /* we need to free temp->image[suffix] too */
          for (l = 0; l <= suffix; l++)
            SDL_FreeSurface(temp->image[l]);
          free(temp->image);
          free(temp);
          return NULL;
        }
        if (SDL_SetColorKey(temp->image[suffix], SDL_SRCCOLORKEY,
                            transparent_color) != 0)
        {
          fprintf(stderr, "tenm_load_image: SDL_SetColorKey failed (%s)\n",
                  SDL_GetError());
          SDL_FreeSurface(master);
          /* we need to free temp->image[suffix] too */
          for (l = 0; l <= suffix; l++)
            SDL_FreeSurface(temp->image[l]);
          free(temp->image);
          free(temp);
          return NULL;
        }
      }
    temp->n += row * column;
    SDL_FreeSurface(master);
  }

  va_end(ap);

  return temp;
}

void
tenm_image_delete(tenm_image *p)
{
  int i;
  /* sanity check */
  if (p == NULL)
    return;

  for (i = 0; i < p->n; i++)
    SDL_FreeSurface(p->image[i]);
  free(p->image);
  free(p);
}

/* return 0 on success, 1 on error */
int
tenm_draw_image(int x, int y, tenm_image *image, int number)
{
  SDL_Rect temp;

  /* sanity check */
  if (image == NULL)
    return 0;
  if ((number < 0) || (number >= image->n))
    return 0;

  temp.x = x;
  temp.y = y;
  temp.w = image->image[number]->w;
  temp.h = image->image[number]->h;

  if (SDL_BlitSurface(image->image[number], NULL, dpy, &temp) != 0)
  {
    fprintf(stderr, "tenm_draw_image: SDL_BlitSurface failed (%s)\n",
            SDL_GetError());
    return 1;
  }

  return 0;
}

/* return 0 on success, 1 on error */
int
tenm_draw_string(int x, int y, tenm_image *font, const char *string,
                 int length)
{
  int i;
  int suffix;
  int temp_x;
  int status = 0;

  /* sanity check */
  if (font == NULL)
    return 0;
  if (font->n < TENM_FONT_NUMBER_CHAR)
    return 0;
  if (string == NULL)
    return 0;
  if (length <= 0)
    return 0;

  temp_x = x;
  for (i = 0; (i < length) && (string[i] != '\0'); i++)
  {
    if (!isprint(string[i]))
    {
      fprintf(stderr, "tenm_draw_string: char %d unprintable (\\x%x)\n",
              i, string[i]);
      status = 1;
      continue;
    }
    suffix = tenm_font_index(string[i]);
    if ((suffix < 0) || (suffix >= font->n))
    {
      fprintf(stderr, "tenm_draw_string: tenm_font_index(%d) returned "
              "strange number (%d)\n", i, suffix);
      status = 1;
      continue;
    }

    if (tenm_draw_image(temp_x, y - font->image[suffix]->h / 2,
                        font, suffix) != 0)
    {
      fprintf(stderr, "tenm_draw_string: tenm_draw_image(%d) failed\n", i);
      status = 1;
      continue;
    }
    
    temp_x += font->image[suffix]->w;
  }
  return status;
}

/* order of characters (ASCII order)
 *    !"#$%&'()*+,-./0123456789:;<=>?
 *   @ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_
 *   `abcdefghijklmnoprqstuvwxyz{|}~
 */

/* return a non-negative index for the character c
 * return -1 on error
 */
static int
tenm_font_index(int c)
{
  /* sanity check */
  if (!isprint(c))
  {
    fprintf(stderr, "tenm_font_index: unprintable char (\\x%x)\n", c);
    return -1;
  }
  
  /* don't assume ASCII */
  switch (c)
  {
  case ' ':
    return 0;
    break;
  case '!':
    return 1;
    break;
  case '"':
    return 2;
    break;
  case '#':
    return 3;
    break;
  case '$':
    return 4;
    break;
  case '%':
    return 5;
    break;
  case '&':
    return 6;
    break;
  case '\'':
    return 7;
    break;
  case '(':
    return 8;
    break;
  case ')':
    return 9;
    break;
  case '*':
    return 10;
    break;
  case '+':
    return 11;
    break;
  case ',':
    return 12;
    break;
  case '-':
    return 13;
    break;
  case '.':
    return 14;
    break;
  case '/':
    return 15;
    break;
  case '0':
    return 16;
    break;
  case '1':
    return 17;
    break;
  case '2':
    return 18;
    break;
  case '3':
    return 19;
    break;
  case '4':
    return 20;
    break;
  case '5':
    return 21;
    break;
  case '6':
    return 22;
    break;
  case '7':
    return 23;
    break;
  case '8':
    return 24;
    break;
  case '9':
    return 25;
    break;
  case ':':
    return 26;
    break;
  case ';':
    return 27;
    break;
  case '<':
    return 28;
    break;
  case '=':
    return 29;
    break;
  case '>':
    return 30;
    break;
  case '?':
    return 31;
    break;
  case '@':
    return 32;
    break;
  case 'A':
    return 33;
    break;
  case 'B':
    return 34;
    break;
  case 'C':
    return 35;
    break;
  case 'D':
    return 36;
    break;
  case 'E':
    return 37;
    break;
  case 'F':
    return 38;
    break;
  case 'G':
    return 39;
    break;
  case 'H':
    return 40;
    break;
  case 'I':
    return 41;
    break;
  case 'J':
    return 42;
    break;
  case 'K':
    return 43;
    break;
  case 'L':
    return 44;
    break;
  case 'M':
    return 45;
    break;
  case 'N':
    return 46;
    break;
  case 'O':
    return 47;
    break;
  case 'P':
    return 48;
    break;
  case 'Q':
    return 49;
    break;
  case 'R':
    return 50;
    break;
  case 'S':
    return 51;
    break;
  case 'T':
    return 52;
    break;
  case 'U':
    return 53;
    break;
  case 'V':
    return 54;
    break;
  case 'W':
    return 55;
    break;
  case 'X':
    return 56;
    break;
  case 'Y':
    return 57;
    break;
  case 'Z':
    return 58;
    break;
  case '[':
    return 59;
    break;
  case '\\':
    return 60;
    break;
  case ']':
    return 61;
    break;
  case '^':
    return 62;
    break;
  case '_':
    return 63;
    break;
  case '`':
    return 64;
    break;
  case 'a':
    return 65;
    break;
  case 'b':
    return 66;
    break;
  case 'c':
    return 67;
    break;
  case 'd':
    return 68;
    break;
  case 'e':
    return 69;
    break;
  case 'f':
    return 70;
    break;
  case 'g':
    return 71;
    break;
  case 'h':
    return 72;
    break;
  case 'i':
    return 73;
    break;
  case 'j':
    return 74;
    break;
  case 'k':
    return 75;
    break;
  case 'l':
    return 76;
    break;
  case 'm':
    return 77;
    break;
  case 'n':
    return 78;
    break;
  case 'o':
    return 79;
    break;
  case 'p':
    return 80;
    break;
  case 'q':
    return 81;
    break;
  case 'r':
    return 82;
    break;
  case 's':
    return 83;
    break;
  case 't':
    return 84;
    break;
  case 'u':
    return 85;
    break;
  case 'v':
    return 86;
    break;
  case 'w':
    return 87;
    break;
  case 'x':
    return 88;
    break;
  case 'y':
    return 89;
    break;
  case 'z':
    return 90;
    break;
  case '{':
    return 91;
    break;
  case '|':
    return 92;
    break;
  case '}':
    return 93;
    break;
  case '~':
    return 94;
    break;
  default:
    fprintf(stderr, "tenm_font_index: unknown char (\\x%x)\n", c);
    return -1;
    break;
  }
  /* should not reach here */
  return -1;
}
